#include <windows.h>
#include <commctrl.h>
#include <iostream>
#include <fstream>
#include <sstream>
#include <string>
#include <cmath>
#include <ctime>
#include <cstring>
#include <cstdlib>

const DWORD WIN_INFINITE = INFINITE;        // Doubly defined in windows.h and apfloat.h

#include "ap.h"
#include "aptest.rh"


using namespace std;


#ifndef M_PI
#define M_PI 3.14159265358979323846
#endif


LRESULT CALLBACK WndProc (HWND, UINT, WPARAM, LPARAM);
LRESULT CALLBACK SetDialog (HWND hdlg, UINT messg, WPARAM wParam, LPARAM lParam);
LRESULT CALLBACK RunDialog (HWND hdlg, UINT messg, WPARAM wParam, LPARAM lParam);
LRESULT CALLBACK AboutDialog (HWND hdlg, UINT messg, WPARAM wParam, LPARAM lParam);

char szProgramName[] = "aptest";
HINSTANCE hInstProg;
HANDLE hThread = 0;
HWND hWnd, hDlg, hProg;
HDC hCDC;
SIZE size;
RECT client;
HBITMAP oldbitmap, newbitmap;

int method = 0, base = 10;
size_t prec = 133000;
bool docalculate = false;
ofstream out;

char *bases[] =
{"2", "3", "4", "5", "6", "7", "8", "9", "10", "11",
 "12", "13", "14", "15", "16", "17", "18", "19", "20", "21",
 "22", "23", "24", "25", "26", "27", "28", "29", "30", "31",
 "32", "33", "34", "35", "36"};

// Program for testing the apfloat class, actually calculates pi

void print (const char *c)
{
    HDC hdc;
    char buf[260];

    BitBlt (hCDC, 0, 0, client.right, client.bottom - size.cy,
            hCDC, 0, size.cy, SRCCOPY);
    strncpy (buf, c, 260);
    memset (buf + strlen (buf), ' ', 260 - strlen (c));         // Bogus

    TextOut (hCDC, 0, client.bottom - size.cy, buf, 260);

    hdc = GetDC (hWnd);
    BitBlt (hdc, 0, 0, client.right, client.bottom, hCDC, 0, 0, SRCCOPY);
    ReleaseDC (hWnd, hdc);
}

void dump (void)
{
    int i;

    { ostringstream s; s << "Ramsize = " << Ramsize; print (s.str ().c_str ()); }
    { ostringstream s; s << "CacheL1size = " << CacheL1size; print (s.str ().c_str ()); }
    { ostringstream s; s << "CacheL2size = " << CacheL2size; print (s.str ().c_str ()); }
    { ostringstream s; s << "Cacheburst = " << Cacheburst; print (s.str ().c_str ()); }
    { ostringstream s; s << "Maxblocksize = " << Maxblocksize; print (s.str ().c_str ()); }
    { ostringstream s; s << "Memorytreshold = " << Memorytreshold; print (s.str ().c_str ()); }
    { ostringstream s; s << "Blocksize = " << Blocksize; print (s.str ().c_str ()); }
    { ostringstream s; s << "Cachetreshold = " << Cachetreshold; print (s.str ().c_str ()); }
    { ostringstream s; s << "Cacheburstblocksize = " << Cacheburstblocksize; print (s.str ().c_str ()); }
    { ostringstream s; s << "Cachemaxblocksize = " << Cachemaxblocksize; print (s.str ().c_str ()); }
    { ostringstream s; s << "Cacheblocksize = " << Cacheblocksize; print (s.str ().c_str ()); }
    { ostringstream s; s << "Base = " << Base; print (s.str ().c_str ()); }
    { ostringstream s; s << "Basedigit = " << Basedigit; print (s.str ().c_str ()); }
    { ostringstream s; s << "Basedigits = " << Basedigits; print (s.str ().c_str ()); }
    { ostringstream s; s << "NBasefactors = " << NBasefactors; print (s.str ().c_str ()); }
    ostringstream s;
    s << "Basefactors = ";

    for (i = 0; i < NBasefactors; i++)
        s << (i ? ", " : "") << Basefactors[i];

    print (s.str ().c_str ());
}

size_t rnd2base (size_t x)
{
    size_t r = 1;

    x /= Basedigits;

    if (!x) return 0;

    while (x >>= 1) r <<= 1;

    return r * Basedigits;
}

size_t rnd3base (size_t x)
{
    size_t r = 3, p = 3;

    if (!x) return 0;

    x /= Basedigits;

    while (r <= x)
    {
        p = r;
        r <<= 1;
    }

    return p * Basedigits;
}

double baseprec4 (int k, int r = 4)
{
    return (pow (4.0, (double) k) * sqrt((double) r) * M_PI - log (16.0 * sqrt ((double) r)) - k * log (4.0)) / log ((double) Basedigit);
}

double baseprec2 (int k, int r = 4)
{
    return (pow (2.0, (double) k) * sqrt((double) r) * M_PI - log (16.0 * sqrt ((double) r)) - k * log (2.0)) / log ((double) Basedigit);
}

// Calculates pi using the Borweins' quartic algorithm

apfloat pi4 (size_t destprec)
{
    int maxk = 0, k;
    size_t prec, dprec;
    time_t b;
    bool h;

    prec = rnd23up (destprec / Basedigits);
    if (prec == (prec & -prec))
        h = false;
    else
        h = true;

    while (baseprec4 (maxk) < destprec) maxk++;

    PostMessage (hProg, PBM_SETRANGE, 0, MAKELPARAM (0, 6 * maxk + 2));
    PostMessage (hProg, PBM_SETSTEP, 1, 0);

    print ("Using the Borweins' quartic iteration");
    { ostringstream s; s << "Total " << maxk << " iterations"; print (s.str ().c_str ()); }

    if (!docalculate) return 0;

    b = time (0);

    apfloat y = sqrt (apfloat (2, destprec)) - 1;
    apfloat a = 2 - 4 * y;
    apfloat half = apfloat (1, destprec) / 2;
    apfloat z;

    { ostringstream s; s << "Initial values took " << difftime (time (0), b) << " seconds"; print (s.str ().c_str ()); }

    if (!docalculate) return 0;

    PostMessage (hProg, PBM_STEPIT, 0, 0);

    for (k = 0; k < maxk; k++)
    {
        b = time (0);

        if (k < 2)
            z = pow (y, 4);
        else
        {
            z = y;
            prec = destprec + 4 * z.exp ();
            z.prec (prec > 1 ? prec : 1);
            z = pow (z, 4);
        }

        y = 1 - z;

        if (!docalculate) return 0;

        PostMessage (hProg, PBM_STEPIT, 0, 0);

        if (k < 5)
            y = invroot (y, 4);
        else
        {
            if (h)
                prec = rnd3base (-2 * z.exp ());
            else
                prec = rnd2base (-2 * z.exp ());
            z.prec (-z.exp ());
            z = 1 + z / 4;
            z.prec (prec);
            y = invroot (y, 4, y.prec (), z, prec);
        }

        if (!docalculate) return 0;

        PostMessage (hProg, PBM_STEPIT, 0, 0);

        if (k < 5)
            y = (y - 1) / (y + 1);
        else
        {
            apfloat t;

            z = y - 1;
            if (h)
                prec = rnd3base (-2 * z.exp ());
            else
                prec = rnd2base (-2 * z.exp ());
            dprec = z.prec ();
            t = z;
            t.prec (-z.exp ());
            t = half - t / 4;
            t.prec (prec);
            y = z * invroot (y + 1, 1, dprec, t, prec);
        }

        if (!docalculate) return 0;

        PostMessage (hProg, PBM_STEPIT, 0, 0);

        z = pow (y + 1, 2);

        if (!docalculate) return 0;

        PostMessage (hProg, PBM_STEPIT, 0, 0);

        a *= z * z;

        if (!docalculate) return 0;

        PostMessage (hProg, PBM_STEPIT, 0, 0);

        a -= (1UL << (2 * k + 3)) * y * (z - y);

        { ostringstream s; s << "Iteration " << k + 1 << " took " << difftime (time (0), b) << " seconds"; print (s.str ().c_str ()); }

        if (!docalculate) return 0;

        PostMessage (hProg, PBM_STEPIT, 0, 0);
    }

    b = time (0);

    a = 1 / a;

    { ostringstream s; s << "Final inverse took " << difftime (time (0), b) << " seconds"; print (s.str ().c_str ()); }

    return a;
}

// Calculates pi using the Gauss-Legendre algorithm

apfloat pi2 (size_t destprec)
{
    int maxk = 0, k;
    size_t prec;
    time_t m;
    bool h;

    prec = rnd23up (destprec / Basedigits);
    if (prec == (prec & -prec))
        h = false;
    else
        h = true;

    while (baseprec2 (maxk) < destprec) maxk++;

    PostMessage (hProg, PBM_SETRANGE, 0, MAKELPARAM (0, 5 * maxk + 2));
    PostMessage (hProg, PBM_SETSTEP, 1, 0);

    print ("Using the Gauss-Legendre iteration");
    { ostringstream s; s << "Total " << maxk << " iterations"; print (s.str ().c_str ()); }

    if (!docalculate) return 0;

    m = time (0);

    apfloat b = invroot (apfloat (2, destprec), 2);
    apfloat a = apfloat (1, destprec), y;
    apfloat t = a / 4, g;

    { ostringstream s; s << "Initial values took " << difftime (time (0), m) << " seconds"; print (s.str ().c_str ()); }

    if (!docalculate) return 0;

    PostMessage (hProg, PBM_STEPIT, 0, 0);

    for (k = 0; k < maxk; k++)
    {
        m = time (0);

        y = a;
        a = (a + b) / 2;

        if (!docalculate) return 0;

        PostMessage (hProg, PBM_STEPIT, 0, 0);

        if (k < 6)
        {
            apfloat z = b * y;

            if (!docalculate) return 0;

            PostMessage (hProg, PBM_STEPIT, 0, 0);

            b = 0;              // Save space
            g = invroot (z, 2);

            if (!docalculate) return 0;

            PostMessage (hProg, PBM_STEPIT, 0, 0);

            b = z * g;
            if (h)
                g.prec (rnd3base ((size_t) (200.0 / log ((double) Basedigit))));
            else
                g.prec (rnd2base ((size_t) (200.0 / log ((double) Basedigit))));
        }
        else
        {
            apfloat z = b * y;

            if (!docalculate) return 0;

            PostMessage (hProg, PBM_STEPIT, 0, 0);

            b = 0;              // Save space
            g = invroot (z, 2, destprec, g);

            if (!docalculate) return 0;

            PostMessage (hProg, PBM_STEPIT, 0, 0);

            b = z * g;
            if (h)
                g.prec (rnd3base ((size_t) (200.0 / log ((double) Basedigit)) << k - 5));
            else
                g.prec (rnd2base ((size_t) (200.0 / log ((double) Basedigit)) << k - 5));
        }

        if (!docalculate) return 0;

        PostMessage (hProg, PBM_STEPIT, 0, 0);

        if (k)
            t -= (1UL << k) * pow (y - a, 2);
        else
            t -= pow (y - a, 2);

        { ostringstream s; s << "Iteration " << k + 1 << " took " << difftime (time (0), m) << " seconds"; print (s.str ().c_str ()); }

        if (!docalculate) return 0;

        PostMessage (hProg, PBM_STEPIT, 0, 0);
    }

    m = time (0);

    a = pow (a + b, 2) / (4 * t);

    { ostringstream s; s << "Final value took " << difftime (time (0), m) << " seconds"; print (s.str ().c_str ()); }

    return a;
}

// Calculates pi using the binary splitting algorithm of the Chudnovsky brothers
// Code is essentially taken from CLN by Bruno Haible

apfloat A, B, J, one, two, five, six;
size_t maxk, currk, oldpct;

apfloat a (size_t n)
{
    apfloat v = A + B * n;

    v.sign (1 - 2 * (n & 1));

    return v;
}

apfloat p (size_t n)
{
    apfloat f = n, sixf = six * f;

    if (!n)
        return one;
    else
        return (sixf - one) * (two * f - one) * (sixf - five);
}

apfloat q (size_t n)
{
    apfloat f = n;

    if (!n)
        return one;
    else
        return J * f * f * f;
}

bool r (size_t N1, size_t N2, apfloat *T, apfloat *Q, apfloat *P)
{
    size_t pct;

    switch (N2 - N1)
    {
        case 0:
        {
            assert (N1 != N2);

            break;
        }
        case 1:
        {
            apfloat p0 = p (N1);

            *T = a (N1) * p0;
            *Q = q (N1);
            if (P) *P = p0;

            currk += 1;

            break;
        }
        case 2:
        {
            apfloat p0 = p (N1), p01 = p0 * p (N1 + 1),
                    q1 = q (N1 + 1);

            *T = q1 * a (N1) * p0 +
                 a (N1 + 1) * p01;
            *Q = q (N1) * q1;
            if (P) *P = p01;

            currk += 4;

            break;
        }
        case 3:
        {
            apfloat p0 = p (N1), p01 = p0 * p (N1 + 1), p012 = p01 * p (N1 + 2),
                    q2 = q (N1 + 2), q12 = q (N1 + 1) * q2;

            *T = q12 * a (N1) * p0 +
                 q2 * a (N1 + 1) * p01 +
                 a (N1 + 2) * p012;
            *Q = q (N1) * q12;
            if (P) *P = p012;

            currk += 8;

            break;
        }
        case 4:
        {
            apfloat p0 = p (N1), p01 = p0 * p (N1 + 1), p012 = p01 * p (N1 + 2), p0123 = p012 * p (N1 + 3),
                    q3 = q (N1 + 3), q23 = q (N1 + 2) * q3, q123 = q (N1 + 1) * q23;

            *T = q123 * a (N1) * p0 +
                 q23 * a (N1 + 1) * p01 +
                 q3 * a (N1 + 2) * p012 +
                 a (N1 + 3) * p0123;
            *Q = q (N1) * q123;
            if (P) *P = p0123;

            currk += 12;

            break;
        }
        default:
        {
            size_t Nm = (N1 + N2) / 2;
            apfloat LT, LQ, LP, RT, RQ, RP;

            if (!docalculate) return false;

            r (N1, Nm, &LT, &LQ, &LP);

            if (!docalculate) return false;

            r (Nm, N2, &RT, &RQ, (P ? &RP : 0));

            if (!docalculate) return false;

            *T = RQ * LT + LP * RT;

            if (!docalculate) return false;

            *Q = LQ * RQ;

            if (!docalculate) return false;

            if (P) *P = LP * RP;

            if (!docalculate) return false;

            currk += N2 - N1;
        }
    }

    pct = (size_t) (100.0 * currk / maxk);

    if (pct != oldpct)
    {
        PostMessage (hProg, PBM_SETPOS, (WPARAM) (10000.0 * currk / maxk), 0);
        Sleep (0);          // Gives the progress bar a time slice to update (mostly needed for on NT)

        oldpct = pct;
    }

    return true;
}

apfloat pib (size_t destprec)
{
    size_t n;
    apfloat T, Q;

    A = 13591409;
    B = 545140134;
    J = 101568000; J *= 107701824;       // J = 10939058860032000

    one = 1;
    two = 2;
    five = 5;
    six = 6;

    A.prec (destprec);
    B.prec (destprec);
    J.prec (destprec);

    one.prec (destprec);
    two.prec (destprec);
    five.prec (destprec);
    six.prec (destprec);

    n = (size_t) (destprec * log ((double) Basedigit) / 32.65445004177);
    maxk = (size_t) (n * (log ((double) n) / log (2.0) + 1));
    currk = 0;
    oldpct = 0;

    print ("Using the Chudnovsky brothers' binary splitting algorithm");

    PostMessage (hProg, PBM_SETRANGE, 0, MAKELPARAM (0, 10000));

    if (!r (0, n + 1, &T, &Q, 0))
        return 0;                       // Cancelled

    print ("Final value");

    return invroot (invroot (apfloat (640320, destprec), 2) * T, 1) * 53360 * Q;
}

DWORD calculate (LPDWORD x)
{
    time_t b;
    apfloat p = 0;

    apbase (base);
    dump ();

    { ostringstream s; s << "Calculating pi to " << prec << " base-" << Basedigit << " digits"; print (s.str ().c_str ()); }

    b = time (0);

    if (method == 0)
        p = pib (prec);
    else if (method == 1)
        p = pi2 (prec);
    else if (method == 2)
        p = pi4 (prec);

    if (p != 0)
    {
        out << pretty << p << endl;
    }
    else
    {
        print ("Cancelled");
    }

    out.close ();

    { ostringstream s; s << "Total time " << difftime (time (0), b) << " seconds"; print (s.str ().c_str ()); }

    PostMessage (hDlg, WM_COMMAND, IDOK, 0);        // Clear calculation dialog

    return 0;
}

int WINAPI WinMain (HINSTANCE hInst, HINSTANCE hPrevInst, LPSTR lpszCmdLine, int nCmdShow)
{
    MSG msg;
    WNDCLASS ownclass;

    hInstProg = hInst;

    if (!hPrevInst)
    {
        ownclass.lpszClassName = szProgramName;
        ownclass.hInstance = hInst;
        ownclass.lpfnWndProc = (WNDPROC) WndProc;
        ownclass.hCursor = LoadCursor (NULL, IDC_ARROW);
        ownclass.hIcon = LoadIcon (hInst, MAKEINTRESOURCE (ID_APICON));
        ownclass.lpszMenuName = MAKEINTRESOURCE (ID_APMENU);
        ownclass.hbrBackground = (HBRUSH) GetStockObject (WHITE_BRUSH);
        ownclass.style = CS_HREDRAW|CS_VREDRAW;
        ownclass.cbClsExtra = 0;
        ownclass.cbWndExtra = 0;

        if (!RegisterClass (&ownclass))
            return FALSE;
    }

    hWnd = CreateWindow (szProgramName, "Aptest",
                         WS_OVERLAPPEDWINDOW, CW_USEDEFAULT, CW_USEDEFAULT,
                         CW_USEDEFAULT, CW_USEDEFAULT, NULL, NULL, hInst, NULL);

    ShowWindow (hWnd, nCmdShow);
    UpdateWindow (hWnd);

    while (GetMessage (&msg, NULL, 0, 0))
    {
        TranslateMessage (&msg);
        DispatchMessage (&msg);
    }

    return TRUE;
}

LRESULT CALLBACK WndProc (HWND hwnd, UINT messg, WPARAM wParam, LPARAM lParam)
{
    PAINTSTRUCT ps;
    HDC hdc;
    LONG w, h;
    OPENFILENAME ofn;
    char filename[260];

    memset (filename, 0, 260);

    switch (messg)
    {
        case WM_CREATE:
            GetClientRect (hwnd, &client);
            w = client.right;
            h = client.bottom;

            hdc = GetDC (hwnd);
            GetTextExtentPoint32 (hdc, "Apfloat", 7, &size);
            hCDC = CreateCompatibleDC (hdc);
            newbitmap = CreateCompatibleBitmap (hdc, w, h);
            oldbitmap = (HBITMAP) SelectObject (hCDC, newbitmap);
            FillRect (hCDC, &client, (HBRUSH) (COLOR_WINDOW + 1));
            ReleaseDC (hwnd, hdc);
            break;

        case WM_SIZE:
            if (!hThread)
            {
                GetClientRect (hwnd, &client);
                w = LOWORD (lParam);
                h = HIWORD (lParam);

                DeleteObject (newbitmap);
                SelectObject (hCDC, oldbitmap);
                DeleteDC (hCDC);

                hdc = GetDC (hwnd);
                hCDC = CreateCompatibleDC (hdc);
                newbitmap = CreateCompatibleBitmap (hdc, w, h);
                oldbitmap = (HBITMAP) SelectObject (hCDC, newbitmap);
                FillRect (hCDC, &client, (HBRUSH) (COLOR_WINDOW + 1));
                ReleaseDC (hwnd, hdc);
            }
            break;

        case WM_PAINT:
            hdc = BeginPaint (hwnd, &ps);
            w = client.right;
            h = client.bottom;
            BitBlt (hdc, 0, 0, w, h, hCDC, 0, 0, SRCCOPY);
            ValidateRect (hwnd, NULL);
            EndPaint (hwnd, &ps);
            break;

        case WM_COMMAND:
            switch (wParam)
            {
                case CM_OPEN:
                    ofn.lStructSize = sizeof (OPENFILENAME);
                    ofn.hwndOwner = hwnd;
                    ofn.lpstrFilter = "Text files (*.txt)\0*.txt\0All files (*.*)\0*.*\0\0";
                    ofn.lpstrCustomFilter = NULL;
                    ofn.nFilterIndex = 0;
                    ofn.lpstrFile = filename;
                    ofn.nMaxFile = 260;
                    ofn.lpstrFileTitle = NULL;
                    ofn.lpstrInitialDir = NULL;
                    ofn.lpstrTitle = NULL;
                    ofn.Flags = OFN_PATHMUSTEXIST | OFN_HIDEREADONLY | OFN_EXPLORER;
                    ofn.nFileOffset = 0;
                    ofn.nFileExtension = 0;
                    ofn.lpstrDefExt = "txt";

                    if (GetSaveFileName (&ofn))
                    {
                        if (out.rdbuf ()->is_open ())
                            out.close ();

                        out.open (ofn.lpstrFile);

                        if (out.fail ())
                        {
                            MessageBox (hwnd, "Error opening file!", "Error!", MB_OK);
                        }
                    }
                    break;

                case CM_EXIT:
                    SendMessage (hwnd, WM_DESTROY, 0, 0);
                    break;

                case CM_SET:
                    DialogBox (hInstProg, MAKEINTRESOURCE (ID_SETDIALOG), hwnd, (DLGPROC) SetDialog);
                    break;

                case CM_README:
                    WinExec ("notepad readme.1st", SW_SHOWDEFAULT);
                    break;

                case CM_ABOUT:
                    DialogBox (hInstProg, MAKEINTRESOURCE (ID_ABOUTDIALOG), hwnd, (DLGPROC) AboutDialog);
                    break;

                case CM_RUN:
                    if (!out.rdbuf ()->is_open ())
                    {
                        MessageBox (hwnd, "No output file selected.", "Error!", MB_OK);
                        break;
                    }

                    if (!docalculate && !hThread)
                    {
                        docalculate = true;
                        InitCommonControls ();
                        DialogBox (hInstProg, MAKEINTRESOURCE (ID_RUNDIALOG), hwnd, (DLGPROC) RunDialog);
                    }
                    break;

                default:
                    break;
            }
            break;

        case WM_DESTROY:
            SetCursor (LoadCursor (NULL, IDC_WAIT));

            // The thread should not be running if we come here, as the run dialog has dismissed
            // so this is a precaution
            docalculate = false;
            if (hThread)
            {
                WaitForSingleObject (hThread, WIN_INFINITE);
                CloseHandle (hThread);
                hThread = 0;
            }

            SetCursor (LoadCursor (NULL, IDC_ARROW));
            PostQuitMessage (0);
            break;

        default:
            return (DefWindowProc (hwnd, messg, wParam, lParam));
    }

    return 0L;
}

LRESULT CALLBACK SetDialog (HWND hdlg, UINT messg, WPARAM wParam, LPARAM lParam)
{
    int t;
    ostringstream s;
    char c[30];

    switch (messg)
    {
        case WM_INITDIALOG:
            for (t = 0; t < 35; t++)
                SendDlgItemMessage (hdlg, ID_BASE, CB_INSERTSTRING, -1, (LPARAM) bases[t]);

            SendDlgItemMessage (hdlg, ID_BASE, CB_SETCURSEL, base - 2, 0);
            s << prec;
            SetDlgItemText (hdlg, ID_DIGITS, s.str ().c_str ());

            if (method == 0)
                CheckRadioButton (hdlg, ID_METHODBS, ID_METHODBQ, ID_METHODBS);
            else if (method == 1)
                CheckRadioButton (hdlg, ID_METHODBS, ID_METHODBQ, ID_METHODGL);
            else if (method == 2)
                CheckRadioButton (hdlg, ID_METHODBS, ID_METHODBQ, ID_METHODBQ);

            break;

        case WM_COMMAND:
            switch (LOWORD (wParam))
            {
                case IDOK:
                    GetDlgItemText (hdlg, ID_BASE, c, 30);
                    if (!(istringstream (string (c, 30)) >> base))
                    {
                        MessageBox (hdlg, "Invalid base.", "Error!", MB_OK);
                        break;
                    }
                    GetDlgItemText (hdlg, ID_DIGITS, c, 30);
                    if (!(istringstream (string (c, 30)) >> prec))
                    {
                        MessageBox (hdlg, "Invalid number of digits.", "Error!", MB_OK);
                        break;
                    }

                    if (IsDlgButtonChecked (hdlg, ID_METHODBS))
                        method = 0;
                    else if (IsDlgButtonChecked (hdlg, ID_METHODGL))
                        method = 1;
                    else if (IsDlgButtonChecked (hdlg, ID_METHODBQ))
                        method = 2;

                    EndDialog (hdlg, wParam);
                    break;

                case IDCANCEL:
                    EndDialog (hdlg, wParam);
                    break;
            }
            break;

        default:
            return FALSE;
    }

    return TRUE;
}

LRESULT CALLBACK RunDialog (HWND hdlg, UINT messg, WPARAM wParam, LPARAM lParam)
{
    RECT rc;
    LONG x;
    DWORD dwThreadId;

    switch (messg)
    {
        case WM_INITDIALOG:
            hDlg = hdlg;
            GetWindowRect (hWnd, &rc);
            x = rc.right;
            GetWindowRect (hdlg, &rc);
            MoveWindow (hdlg, x - rc.right + rc.left, rc.top , rc.right - rc.left, rc.bottom - rc.top, 1);
            hProg = CreateWindowEx (0, PROGRESS_CLASS, "", WS_CHILD | WS_VISIBLE, 16, 48, 256, 20, hDlg, (HMENU) ID_PROGRESS, hInstProg, NULL);
            hThread = CreateThread (NULL, 0, (LPTHREAD_START_ROUTINE) calculate, 0, 0, &dwThreadId);
            break;

        case WM_COMMAND:
            switch (LOWORD (wParam))
            {
                case IDOK:
                case IDCANCEL:
                    docalculate = false;
                    if (hThread)
                    {
                        WaitForSingleObject (hThread, WIN_INFINITE);
                        CloseHandle (hThread);
                        hThread = 0;
                    }
                    EndDialog (hdlg, wParam);
                    break;
            }
            break;

        default:
            return FALSE;
    }

    return TRUE;
}

LRESULT CALLBACK AboutDialog (HWND hdlg, UINT messg, WPARAM wParam, LPARAM lParam)
{
    switch (messg)
    {
        case WM_INITDIALOG:
            break;

        case WM_COMMAND:
            switch (LOWORD (wParam))
            {
                case IDOK:
                case IDCANCEL:
                    EndDialog (hdlg, wParam);
                    break;
            }
            break;

        default:
            return FALSE;
    }

    return TRUE;
}
